/*
MIT License

Copyright (c) 2009 Justin (tuxinator2009) Davis

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <SDL.h>
#include <SDL_image.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <map>
#include <vector>
#include <list>
#include "dungeongenerator.h"
#include "randomnumbergenerator.h"

Point& Point::operator=(const Point &other)
{
	x = other.x;
	y = other.y;
	count = other.count;
	return *this;
}

bool Point::operator==(const Point &other) const
{
	if (x == other.x && y == other.y)
		return true;
	return false;
}

bool Point::operator<(const Point &other) const
{
	if (y < other.y)
		return true;
	else if (y == other.y && x < other.x)
		return true;
	return false;
}

bool Point::operator>(const Point &other) const
{
	if (y > other.y)
		return true;
	else if (y == other.y && x > other.x)
		return true;
	return false;
}

DungeonGenerator::DungeonGenerator()
{
	mapData = NULL;
	numRooms = 0;
	roomsPlaced = 0;
	minRoomWidth = 0;
	minRoomHeight = 0;
	maxRoomWidth = 0;
	maxRoomHeight = 0;
	minHallLength = 0;
	maxHallLength = 0;
	mapWidth = 0;
	mapHeight = 0;
	debugMode = false;
	numbers = IMG_Load("numbers.png");
	if (numbers == NULL)
	{
		printf("Error: failed to load numbers.png\n");
		printf("Reason: %s\n", SDL_GetError());
		exit(-1);
	}
	SDL_SetColorKey(numbers, SDL_SRCCOLORKEY, SDL_MapRGB(numbers->format, 255, 0, 255));
}

DungeonGenerator::~DungeonGenerator()
{
	if (mapData != NULL)
	{
		delete[] mapData;
		mapData = NULL;
	}
	if (numbers != NULL)
	{
		SDL_FreeSurface(numbers);
		numbers = NULL;
	}
	generatedMap.clear();
}

void DungeonGenerator::setNumRooms(int value)
{
	numRooms = value;
}

void DungeonGenerator::setMinRoomWidth(int value)
{
	minRoomWidth = value;
}

void DungeonGenerator::setMinRoomHeight(int value)
{
	minRoomHeight = value;
}

void DungeonGenerator::setMaxRoomWidth(int value)
{
	maxRoomWidth = value;
}

void DungeonGenerator::setMaxRoomHeight(int value)
{
	maxRoomHeight = value;
}

void DungeonGenerator::setMinHallLength(int value)
{
	minHallLength = value;
}

void DungeonGenerator::setMaxHallLength(int value)
{
	maxHallLength = value;
}

void DungeonGenerator::generateDungeon(SDL_Surface *screen, unsigned long seed, bool debug)
{
	Point point, point2;
	unsigned long previousTime, currentTime, updateTime = UPDATE_TIME;
	int roomWidth, roomHeight, hallLength, availableDirections, randomNumber, roomPlacementAttempts, wallIndex;
	int maxPlacementAttempts = MAX_PLACEMENT_ATTEMPTS;
	bool north, south, east, west, done;
	debugMode = debug;
	numberGenerator.seedGenerator(seed);
	generatedMap.clear();
	wallList.clear();
	roomsPlaced = 0;
	roomWidth = (int)(numberGenerator.getRandomNumber((unsigned long)(minRoomWidth * 1000), (unsigned long)(maxRoomWidth * 1000)) / 1000ul) + 2;
	roomHeight = (int)(numberGenerator.getRandomNumber((unsigned long)(minRoomHeight * 1000), (unsigned long)(maxRoomHeight * 1000)) / 1000ul) + 2;
	addArea(0, 0, roomWidth, roomHeight, DIR_NORTH | DIR_SOUTH | DIR_EAST | DIR_WEST, SPACE_ROOMFLOOR);
	//displayDungeon(screen);
	previousTime = SDL_GetTicks();
	++roomsPlaced;
	while (roomsPlaced < numRooms)
	{
		int hallsToPlace = (int)(numberGenerator.getRandomNumber(0ul, (unsigned long)(HALLS_TO_PLACE * 1000ul)) / 1000);
		done = false;
		if (roomPlacementAttempts < maxPlacementAttempts)
			done = true;
		for (int i = 0; i < hallsToPlace && !done; ++i)
		{
			while (!done) // NOTE: Place Corridor
			{
				hallLength = (int)(numberGenerator.getRandomNumber((unsigned long)(minHallLength * 1000), (unsigned long)(maxHallLength * 1000)) / 1000);
				wallIndex = getRandomPoint();
				point = wallList[wallIndex];
				availableDirections = 0;
				north = false;
				south = false;
				east = false;
				west = false;
				if (isSpaceAvailable(point.x - 1, point.y - hallLength, 3, hallLength))
				{
					north = true;
					++availableDirections;
				}
				if (isSpaceAvailable(point.x - 1, point.y + 1, 3, hallLength))
				{
					south = true;
					++availableDirections;
				}
				if (isSpaceAvailable(point.x + 1, point.y - 1, hallLength, 3))
				{
					east = true;
					++availableDirections;
				}
				if (isSpaceAvailable(point.x - hallLength, point.y - 1, hallLength, 3))
				{
					west = true;
					++availableDirections;
				}
				if (availableDirections > 0)
				{
					randomNumber = (int)(numberGenerator.getRandomNumber(1000ul, (unsigned long)(availableDirections * 1000)) / 1000);
					if (north)
					{
						--randomNumber;
						if (randomNumber == 0)
						{
							removeWall(wallIndex);
							addArea(point.x - 1, point.y - hallLength, 3, hallLength, DIR_NORTH | DIR_EAST | DIR_WEST, SPACE_HALLFLOOR);
							north = false;
							south = false;
							east = false;
							west = false;
							done = true;
						}
					}
					if (south && randomNumber > 0)
					{
						--randomNumber;
						if (randomNumber == 0)
						{
							removeWall(wallIndex);
							addArea(point.x - 1, point.y + 1, 3, hallLength, DIR_SOUTH | DIR_EAST | DIR_WEST, SPACE_HALLFLOOR);
							north = false;
							south = false;
							east = false;
							west = false;
							done = true;
						}
					}
					if (east && randomNumber > 0)
					{
						--randomNumber;
						if (randomNumber == 0)
						{
							removeWall(wallIndex);
							addArea(point.x + 1, point.y - 1, hallLength, 3, DIR_NORTH | DIR_SOUTH | DIR_EAST, SPACE_HALLFLOOR);
							north = false;
							south = false;
							east = false;
							west = false;
							done = true;
						}
					}
					if (west && randomNumber > 0)
					{
						--randomNumber;
						if (randomNumber == 0)
						{
							removeWall(wallIndex);
							addArea(point.x - hallLength, point.y - 1, hallLength, 3, DIR_NORTH | DIR_SOUTH | DIR_WEST, SPACE_HALLFLOOR);
							north = false;
							south = false;
							east = false;
							west = false;
							done = true;
						}
					}
				}
				else
				{
					++wallList[wallIndex].count;
					if (wallList[wallIndex].count == WALL_ATTEMPTS)
					{
						wallList.erase(wallList.begin() + wallIndex);
						generatedMap[point] = SPACE_REMOVEDWALL;
						++numRemovedWalls;
					}
				}
			}
			done = false;
		}
		//displayDungeon(screen);
		done = false;
		roomPlacementAttempts = 0;
		//maxPlacementAttempts = roomsPlaced / 10;
		//if (maxPlacementAttempts < 10)
		//	maxPlacementAttempts = 10;
		while (!done && roomPlacementAttempts < maxPlacementAttempts)
		{
			int max = 0;
			wallIndex = getRandomPoint();
			point = wallList[wallIndex];
			north = false;
			south = false;
			east = false;
			west = false;
			roomWidth = (int)(numberGenerator.getRandomNumber((unsigned long)(minRoomWidth * 1000), (unsigned long)(maxRoomWidth * 1000)) / 1000ul) + 2;
			roomHeight = (int)(numberGenerator.getRandomNumber((unsigned long)(minRoomHeight * 1000), (unsigned long)(maxRoomHeight * 1000)) / 1000ul) + 2;
			point2.x = point.x;
			point2.y = point.y - 1;
			if (generatedMap.find(point2) == generatedMap.end())
				north = true;
			point2.x = point.x;
			point2.y = point.y + 1;
			if (generatedMap.find(point2) == generatedMap.end())
				south = true;
			point2.x = point.x + 1;
			point2.y = point.y;
			if (generatedMap.find(point2) == generatedMap.end())
				east = true;
			point2.x = point.x - 1;
			point2.y = point.y;
			if (generatedMap.find(point2) == generatedMap.end())
				west = true;
			if (north)
				++max;
			if (south)
				++max;
			if (east)
				++max;
			if (west)
				++max;
			if (max == 0)
			{
				++wallList[wallIndex].count;
				if (wallList[wallIndex].count == WALL_ATTEMPTS)
				{
					wallList.erase(wallList.begin() + wallIndex);
					generatedMap[point] = SPACE_REMOVEDWALL;
					++numRemovedWalls;
				}
			}
			randomNumber = (int)(numberGenerator.getRandomNumber(1000ul, (unsigned long)(max * 1000)) / 1000ul);
			if (north)
			{
				--randomNumber;
				if (randomNumber == 0)
				{
					max = 0;
					bool spaceAvailable[roomWidth - 2];
					for (int i = 0; i < roomWidth - 2; ++i)
					{
						spaceAvailable[i] = isSpaceAvailable(point.x - i - 1, point.y - roomHeight, roomWidth, roomHeight);
						if (spaceAvailable[i])
							++max;
					}
					if (max > 0)
					{
						randomNumber = (int)(numberGenerator.getRandomNumber(1000ul, (unsigned long)(max * 1000)) / 1000ul);
						for (int i = 0; i < roomWidth - 2 && randomNumber > 0; ++i)
						{
							if (spaceAvailable[i])
							{
								--randomNumber;
								if (randomNumber == 0)
								{
									addArea(point.x - i - 1, point.y - roomHeight, roomWidth, roomHeight, DIR_NORTH | DIR_SOUTH | DIR_EAST | DIR_WEST, SPACE_ROOMFLOOR);
									removeWall(wallIndex);
									point2.x = point.x;
									point2.y = point.y - 1;
									removeWall(point2);
									++roomsPlaced;
									roomPlacementAttempts = maxPlacementAttempts;
									done = true;
									//printf("(%d/%d) %d%%\n", roomsPlaced, numRooms, (roomsPlaced * 100) / numRooms);
								}
							}
						}
					}
					north = false;
					south = false;
					east = false;
					west = false;
				}
			}
			if (south)
			{
				--randomNumber;
				if (randomNumber == 0)
				{
					max = 0;
					bool spaceAvailable[roomWidth - 2];
					for (int i = 0; i < roomWidth - 2; ++i)
					{
						spaceAvailable[i] = isSpaceAvailable(point.x - i - 1, point.y + 1, roomWidth, roomHeight);
						if (spaceAvailable[i])
							++max;
					}
					if (max > 0)
					{
						randomNumber = (int)(numberGenerator.getRandomNumber(1000ul, (unsigned long)(max * 1000)) / 1000ul);
						for (int i = 0; i < roomWidth - 2 && randomNumber > 0; ++i)
						{
							if (spaceAvailable[i])
							{
								--randomNumber;
								if (randomNumber == 0)
								{
									addArea(point.x - i - 1, point.y + 1, roomWidth, roomHeight, DIR_NORTH | DIR_SOUTH | DIR_EAST | DIR_WEST, SPACE_ROOMFLOOR);
									removeWall(wallIndex);
									point2.x = point.x;
									point2.y = point.y + 1;
									removeWall(point2);
									++roomsPlaced;
									roomPlacementAttempts = maxPlacementAttempts;
									done = true;
									//printf("(%d/%d) %d%%\n", roomsPlaced, numRooms, (roomsPlaced * 100) / numRooms);
								}
							}
						}
					}
					north = false;
					south = false;
					east = false;
					west = false;
				}
			}
			if (east)
			{
				--randomNumber;
				if (randomNumber == 0)
				{
					max = 0;
					bool spaceAvailable[roomHeight - 2];
					for (int i = 0; i < roomHeight - 2; ++i)
					{
						spaceAvailable[i] = isSpaceAvailable(point.x + 1, point.y - i - 1, roomWidth, roomHeight);
						if (spaceAvailable[i])
							++max;
					}
					if (max > 0)
					{
						randomNumber = (int)(numberGenerator.getRandomNumber(1000ul, (unsigned long)(max * 1000)) / 1000ul);
						for (int i = 0; i < roomHeight - 2 && randomNumber > 0; ++i)
						{
							if (spaceAvailable[i])
							{
								--randomNumber;
								if (randomNumber == 0)
								{
									addArea(point.x + 1, point.y - i - 1, roomWidth, roomHeight, DIR_NORTH | DIR_SOUTH | DIR_EAST | DIR_WEST, SPACE_ROOMFLOOR);
									removeWall(wallIndex);
									point2.x = point.x + 1;
									point2.y = point.y;
									removeWall(point2);
									++roomsPlaced;
									roomPlacementAttempts = maxPlacementAttempts;
									done = true;
									//printf("(%d/%d) %d%%\n", roomsPlaced, numRooms, (roomsPlaced * 100) / numRooms);
								}
							}
						}
					}
					north = false;
					south = false;
					east = false;
					west = false;
				}
			}
			if (west)
			{
				--randomNumber;
				if (randomNumber == 0)
				{
					max = 0;
					bool spaceAvailable[roomHeight - 2];
					for (int i = 0; i < roomHeight - 2; ++i)
					{
						spaceAvailable[i] = isSpaceAvailable(point.x - roomWidth, point.y - i - 1, roomWidth, roomHeight);
						if (spaceAvailable[i])
							++max;
					}
					if (max > 0)
					{
						randomNumber = (int)(numberGenerator.getRandomNumber(1000ul, (unsigned long)(max * 1000)) / 1000ul);
						for (int i = 0; i < roomHeight - 2 && randomNumber > 0; ++i)
						{
							if (spaceAvailable[i])
							{
								--randomNumber;
								if (randomNumber == 0)
								{
									addArea(point.x - roomWidth, point.y - i - 1, roomWidth, roomHeight, DIR_NORTH | DIR_SOUTH | DIR_EAST | DIR_WEST, SPACE_ROOMFLOOR);
									removeWall(wallIndex);
									point2.x = point.x - 1;
									point2.y = point.y;
									removeWall(point2);
									++roomsPlaced;
									roomPlacementAttempts = maxPlacementAttempts;
									done = true;
									//printf("(%d/%d) %d%%\n", roomsPlaced, numRooms, (roomsPlaced * 100) / numRooms);
								}
							}
						}
					}
					north = false;
					south = false;
					east = false;
					west = false;
				}
			}
			++roomPlacementAttempts;
		}
		currentTime = SDL_GetTicks();
		if (currentTime - previousTime >= updateTime)
		{
			//SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
			//displayDungeon(screen);
			//displayPercentage(screen);
			//SDL_Flip(screen);
			previousTime += updateTime;
		}
		processInput();
		//SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
		//displayDungeon(screen);
		//displayPercentage(screen);
		//SDL_Flip(screen);
		//SDL_Delay(500);
	}
	SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
	displayPercentage(screen);
	SDL_Flip(screen);
	cleanDungeon(screen);
	SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
	displayDungeon(screen);
	displayPercentage(screen);
	SDL_Flip(screen);
	printf("numRemovedWalls: %d\n", numRemovedWalls);
}

void DungeonGenerator::displayDungeon(SDL_Surface *screen)
{
	SDL_Rect tmp_dst;
	int screenXOffset, screenYOffset, tileSize;
	createMapData();
	tileSize = screen->w / mapWidth;
	if (screen->h / mapHeight < tileSize)
		tileSize = screen->h / mapHeight;
	if (tileSize < MIN_TILE_SIZE)
		tileSize = MIN_TILE_SIZE;
	screenXOffset = (screen->w - (mapWidth * tileSize)) / 2;
	screenYOffset = (screen->h - (mapHeight * tileSize)) / 2;
	for (int y = 0; y < mapHeight; ++y)
	{
		for (int x = 0; x < mapWidth; ++x)
		{
			int location = y * mapWidth + x;
			tmp_dst.x = (x * tileSize) + screenXOffset;
			tmp_dst.y = (y * tileSize) + screenYOffset;
			tmp_dst.w = tileSize;
			tmp_dst.h = tileSize;
			if (mapData[location] == SPACE_UNUSED)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 64, 64, 64));
			else if (mapData[location] == SPACE_WALL)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 192, 192, 192));
			else if (mapData[location] == SPACE_REMOVEDWALL)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 0, 0, 255));
			else if (mapData[location] == SPACE_HALLFLOOR)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 0, 255, 0));
			else if (mapData[location] == SPACE_ROOMFLOOR)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 255, 0, 0));
			else
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 255, 0, 255));
		}
	}
}

void DungeonGenerator::displayDungeon(SDL_Surface *screen, int screenXOffset, int screenYOffset)
{
	SDL_Rect tmp_dst;
	int tileSize;
	tileSize = screen->w / mapWidth;
	if (screen->h / mapHeight < tileSize)
		tileSize = screen->h / mapHeight;
	if (tileSize < MIN_TILE_SIZE)
		tileSize = MIN_TILE_SIZE;
	SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
	for (int y = 0; y < mapHeight; ++y)
	{
		for (int x = 0; x < mapWidth; ++x)
		{
			int location = y * mapWidth + x;
			tmp_dst.x = (x * tileSize) + screenXOffset;
			tmp_dst.y = (y * tileSize) + screenYOffset;
			tmp_dst.w = tileSize;
			tmp_dst.h = tileSize;
			if (mapData[location] == SPACE_UNUSED || mapData[location] == SPACE_WALL || mapData[location] == SPACE_REMOVEDWALL)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 64, 64, 64));
			//else if (mapData[location] == SPACE_WALL)
			//	SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 192, 192, 192));
			//else if (mapData[location] == SPACE_REMOVEDWALL)
			//	SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 0, 0, 255));
			else if (mapData[location] == SPACE_HALLFLOOR)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 0, 255, 0));
			else if (mapData[location] == SPACE_ROOMFLOOR)
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 255, 0, 0));
			else
				SDL_FillRect(screen, &tmp_dst, SDL_MapRGB(screen->format, 255, 0, 255));
		}
	}
	displayPercentage(screen);
	SDL_Flip(screen);
}

int DungeonGenerator::getMapWidth()
{
	return mapWidth;
}

int DungeonGenerator::getMapHeight()
{
	return mapHeight;
}

int DungeonGenerator::getRandomPoint()
{
	unsigned int randomNumber;
	randomNumber = (unsigned int)numberGenerator.getRandomNumber(1ul, (unsigned long)wallList.size()) - 1;
	return randomNumber;
}

bool DungeonGenerator::isSpaceAvailable(int x, int y, int w, int h)
{
	for (int y2 = 0; y2 < h; ++y2)
	{
		for (int x2 = 0; x2 < w; ++x2)
		{
			Point point;
			point.x = x2 + x;
			point.y = y2 + y;
			if (generatedMap.find(point) != generatedMap.end())
				return false;
		}
	}
	return true;
}

bool DungeonGenerator::isAdjacentToFloor(Point point)
{
	Point point2;
	int type;
	point2.x = point.x;
	point2.y = point.y - 1;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		type = generatedMap[point2];
		if (type == SPACE_HALLFLOOR || type == SPACE_ROOMFLOOR)
			return true;
	}
	point2.x = point.x - 1;
	point2.y = point.y;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		type = generatedMap[point2];
		if (type == SPACE_HALLFLOOR || type == SPACE_ROOMFLOOR)
			return true;
	}
	point2.x = point.x + 1;
	point2.y = point.y;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		type = generatedMap[point2];
		if (type == SPACE_HALLFLOOR || type == SPACE_ROOMFLOOR)
			return true;
	}
	point2.x = point.x;
	point2.y = point.y + 1;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		type = generatedMap[point2];
		if (type == SPACE_HALLFLOOR || type == SPACE_ROOMFLOOR)
			return true;
	}
	return false;
}

void DungeonGenerator::addArea(int x, int y, int w, int h, int surroundSides, int floorType)
{
	for (int y2 = 0; y2 < h; ++y2)
	{
		for (int x2 = 0; x2 < w; ++x2)
		{
			Point point;
			point.x = x2 + x;
			point.y = y2 + y;
			if ((surroundSides & DIR_NORTH) != 0 && y2 == 0)
				generatedMap.insert(std::make_pair(point, SPACE_WALL));
			else if ((surroundSides & DIR_SOUTH) != 0 && y2 == h - 1)
				generatedMap.insert(std::make_pair(point, SPACE_WALL));
			else if ((surroundSides & DIR_EAST) != 0 && x2 == w - 1)
				generatedMap.insert(std::make_pair(point, SPACE_WALL));
			else if ((surroundSides & DIR_WEST) != 0 && x2 == 0)
				generatedMap.insert(std::make_pair(point, SPACE_WALL));
			else
				generatedMap.insert(std::make_pair(point, floorType));
		}
	}
	for (int y2 = 0; y2 < h; ++y2)
	{
		for (int x2 = 0; x2 < w; ++x2)
		{
			Point point;
			point.x = x2 + x;
			point.y = y2 + y;
			if (generatedMap[point] == SPACE_WALL && numNeighbors(point, floorType) > 0)
				wallList.push_back(point);
		}
	}
}

void DungeonGenerator::createMapData()
{
	int minX = 0;
	int maxX = 0;
	int minY = 0;
	int maxY = 0;
	int offsetX, offsetY;
	for (iter = generatedMap.begin(); iter != generatedMap.end(); ++iter)
	{
		if (iter->first.x < minX)
			minX = iter->first.x;
		if (iter->first.x > maxX)
			maxX = iter->first.x;
		if (iter->first.y < minY)
			minY = iter->first.y;
		if (iter->first.y > maxY)
			maxY = iter->first.y;
	}
	mapWidth = maxX - minX + 5;
	mapHeight = maxY - minY + 5;
	offsetX = (-1 * minX) + 2;
	offsetY = (-1 * minY) + 2;
	if (mapData != NULL)
	{
		delete[] mapData;
		mapData = NULL;
	}
	mapData = new int[mapWidth * mapHeight];
	for (int y = 0; y < mapHeight; ++y)
	{
		for (int x = 0; x < mapWidth; ++x)
		{
			Point point;
			point.x = x - offsetX;
			point.y = y - offsetY;
			if (generatedMap.find(point) == generatedMap.end())
				mapData[y * mapWidth + x] = SPACE_UNUSED;
			else
				mapData[y * mapWidth + x] = generatedMap[point];
		}
	}
}

int DungeonGenerator::numNeighbors(Point point, int type)
{
	Point point2;
	int numNeighbors = 0;
	point2.x = point.x;
	point2.y = point.y - 1;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		if (generatedMap[point2] == type)
			++numNeighbors;
	}
	point2.x = point.x - 1;
	point2.y = point.y;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		if (generatedMap[point2] == type)
			++numNeighbors;
	}
	point2.x = point.x + 1;
	point2.y = point.y;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		if (generatedMap[point2] == type)
			++numNeighbors;
	}
	point2.x = point.x;
	point2.y = point.y + 1;
	if (generatedMap.find(point2) != generatedMap.end())
	{
		if (generatedMap[point2] == type)
			++numNeighbors;
	}
	return numNeighbors;
}

void DungeonGenerator::cleanDungeon(SDL_Surface *screen)
{
	std::list<Point> pointList;
	unsigned long previousTime, currentTime, updateTime = UPDATE_TIME;
	bool done = false;
	previousTime = SDL_GetTicks();
	for (iter = generatedMap.begin(); iter != generatedMap.end(); ++iter)
	{
		if (iter->second == SPACE_HALLFLOOR && (numNeighbors(iter->first, SPACE_HALLFLOOR) + numNeighbors(iter->first, SPACE_ROOMFLOOR)) <= 1)
			pointList.push_back(iter->first);
	}
	while (pointList.size() > 0)
	{
		for (int i = 0, count = pointList.size(); i < count; ++i)
		{
			Point point;
			generatedMap.erase(pointList.front());
			point.x = pointList.front().x;
			point.y = pointList.front().y - 1;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
				else if (iter->second == SPACE_HALLFLOOR && (numNeighbors(iter->first, SPACE_HALLFLOOR) + numNeighbors(iter->first, SPACE_ROOMFLOOR)) <= 1)
					pointList.push_back(point);
			}
			point.x = pointList.front().x - 1;
			point.y = pointList.front().y;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
				else if (iter->second == SPACE_HALLFLOOR && (numNeighbors(iter->first, SPACE_HALLFLOOR) + numNeighbors(iter->first, SPACE_ROOMFLOOR)) <= 1)
					pointList.push_back(point);
			}
			point.x = pointList.front().x + 1;
			point.y = pointList.front().y;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
				else if (iter->second == SPACE_HALLFLOOR && (numNeighbors(iter->first, SPACE_HALLFLOOR) + numNeighbors(iter->first, SPACE_ROOMFLOOR)) <= 1)
					pointList.push_back(point);
			}
			point.x = pointList.front().x;
			point.y = pointList.front().y + 1;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
				else if (iter->second == SPACE_HALLFLOOR && (numNeighbors(iter->first, SPACE_HALLFLOOR) + numNeighbors(iter->first, SPACE_ROOMFLOOR)) <= 1)
					pointList.push_back(point);
			}
			point.x = pointList.front().x - 1;
			point.y = pointList.front().y - 1;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
			}
			point.x = pointList.front().x + 1;
			point.y = pointList.front().y - 1;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
			}
			point.x = pointList.front().x - 1;
			point.y = pointList.front().y + 1;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
			}
			point.x = pointList.front().x + 1;
			point.y = pointList.front().y + 1;
			iter = generatedMap.find(point);
			if (iter != generatedMap.end())
			{
				if (iter->second == SPACE_WALL || iter->second == SPACE_REMOVEDWALL)
					generatedMap.erase(iter);
			}
			pointList.pop_front();
			processInput();
		}
		currentTime = SDL_GetTicks();
		if (currentTime - previousTime >= updateTime)
		{
			//SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
			//displayDungeon(screen);
			//displayPercentage(screen);
			//SDL_Flip(screen);
			previousTime += updateTime;
		}
		//SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
		//displayDungeon(screen);
		//displayPercentage(screen);
		//SDL_Flip(screen);
		processInput();
	}
}

void DungeonGenerator::removeWall(Point point)
{
	bool found = false;
	if (generatedMap.find(point) != generatedMap.end())
	{
		generatedMap[point] = SPACE_HALLFLOOR;
		for (unsigned int i = 0; i < wallList.size() && !found; ++i)
		{
			if (wallList[i] == point)
				wallList.erase(wallList.begin() + i);
		}
	}
}

void DungeonGenerator::removeWall(int index)
{
	if (generatedMap.find(wallList[index]) != generatedMap.end())
	{
		generatedMap[wallList[index]] = SPACE_HALLFLOOR;
		wallList.erase(wallList.begin() + index);
	}
}

void DungeonGenerator::displayPercentage(SDL_Surface *screen)
{
	SDL_Rect tmp_src, tmp_dst;
	int offset;
	char temp_string[128];
	sprintf(temp_string, "%d/%d (%d%%)", roomsPlaced, numRooms, (roomsPlaced * 100) / numRooms);
	offset = (screen->w - (strlen(temp_string) * 32)) / 2;
	for (int i = 0; temp_string[i] != '\0'; ++i)
	{
		if (temp_string[i] >= '0' && temp_string[i] <= '9')
			tmp_src.x = 32 * (temp_string[i] - 48);
		else if (temp_string[i] == '%')
			tmp_src.x = 320;
		else if (temp_string[i] == '/')
			tmp_src.x = 352;
		else if (temp_string[i] == '(')
			tmp_src.x = 384;
		else if (temp_string[i] == ')')
			tmp_src.x = 416;
		else
			tmp_src.x = -1;
		if (tmp_src.x != -1)
		{
			tmp_src.y = 0;
			tmp_src.w = 32;
			tmp_src.h = 32;
			tmp_dst.x = offset + (32 * i);
			tmp_dst.y = screen->h - 32;
			SDL_BlitSurface(numbers, &tmp_src, screen, &tmp_dst);
		}
	}
}

void DungeonGenerator::processInput()
{
	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		if (event.type == SDL_KEYDOWN)
		{
			if (event.key.keysym.sym == SDLK_ESCAPE)
				exit(0);
		}
		else if (event.type == SDL_QUIT)
			exit(0);
	}
}
