/*
MIT License

Copyright (c) 2009 Justin (tuxinator2009) Davis

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <SDL.h>
#include <SDL_image.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <string>
#include "dungeongenerator.h"

int main(int argc, char *argv[])
{
	DungeonGenerator dungeonGenerator;
	SDL_Surface *screen;
	SDL_Event event;
	unsigned long randomSeed, startTime, endTime;
	float frameTime, moveSpeed = 200.0;
	float screenXOffset = 0.0;
	float screenYOffset = 0.0;
	int numRooms, minRoomWidth, minRoomHeight, maxRoomWidth, maxRoomHeight, minHallLength, maxHallLength;
	int mapWidth, mapHeight;
	int tileSize;
	bool quit = false;
	bool done = false;
	bool upPressed = false;
	bool downPressed = false;
	bool leftPressed = false;
	bool rightPressed = false;
	if (argc < 8)
	{
		printf("usage: %s numRooms(3+) minRoomWidth(2-10) minRoomHeight(2-10) maxRoomWidth(2-10) maxRoomHeight(2-10) minHallLength(10-20) maxHallLength(10-20) <randomSeed>\n", argv[0]);
		return -1;
	}
	if (SDL_Init(SDL_INIT_VIDEO) != 0)
	{
		printf("Error: failed to initialize SDL.\n");
		printf("Reason: %s\n", SDL_GetError());
		return -1;
	}
	atexit(SDL_Quit);
	screen = SDL_SetVideoMode(800, 600, 24, SDL_HWSURFACE | SDL_DOUBLEBUF);
	if (screen == NULL)
	{
		printf("Error: failed to setup an 800x600 24bpp screen in windowed mode.\n");
		printf("Reason: %s\n", SDL_GetError());
		return -1;
	}
	numRooms = atoi(argv[1]);
	minRoomWidth = atoi(argv[2]);
	minRoomHeight = atoi(argv[3]);
	maxRoomWidth = atoi(argv[4]);
	maxRoomHeight = atoi(argv[5]);
	minHallLength = atoi(argv[6]);
	maxHallLength = atoi(argv[7]);
	if (numRooms < 3)
	{
		printf("Error: numRooms must be greater than 3.\n");
		return -1;
	}
	if (minRoomWidth < 2 || minRoomWidth > 10)
	{
		printf("Error: minRoomWidth must be between 2 and 10.\n");
		return -1;
	}
	if (minRoomHeight < 2 || minRoomHeight > 10)
	{
		printf("Error: minRoomHeight must be between 2 and 10.\n");
		return -1;
	}
	if (maxRoomWidth < 2 || maxRoomWidth > 10)
	{
		printf("Error: maxRoomWidth must be between 2 and 10.\n");
		return -1;
	}
	if (maxRoomHeight < 2 || maxRoomHeight > 10)
	{
		printf("Error: maxRoomHeight must be between 2 and 10.\n");
		return -1;
	}
	if (minRoomWidth > maxRoomWidth)
	{
		printf("Error: minRoomWidth must be less than maxRoomWidth.\n");
		return -1;
	}
	if (minRoomHeight > maxRoomHeight)
	{
		printf("Error: minRoomHeight must be less than maxRoomHeight.\n");
		return -1;
	}
	if (minHallLength < 10 || minHallLength > 20)
	{
		printf("Error: minHallLength must be between 10 and 20.\n");
		return -1;
	}
	if (maxHallLength < 10 || maxHallLength > 20)
	{
		printf("Error: maxHallLength must be between 10 and 20.\n");
		return -1;
	}
	if (minHallLength > maxHallLength)
	{
		printf("Error: minHallLength must be greater than maxHallLength.\n");
		return -1;
	}
	dungeonGenerator.setNumRooms(numRooms);
	dungeonGenerator.setMinRoomWidth(minRoomWidth);
	dungeonGenerator.setMinRoomHeight(minRoomHeight);
	dungeonGenerator.setMaxRoomWidth(maxRoomWidth);
	dungeonGenerator.setMaxRoomHeight(maxRoomHeight);
	dungeonGenerator.setMinHallLength(minHallLength);
	dungeonGenerator.setMaxHallLength(maxHallLength);
	if (argc == 9)
		randomSeed = strtoul(argv[8], NULL, 10);
	while (!quit)
	{
		while (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_ESCAPE)
						quit = true;
					else if (event.key.keysym.sym == SDLK_RETURN)
					{
						done = false;
						if (argc == 8)
						{
							randomSeed = (unsigned long)time(NULL);
							printf("Random seed: %lu\n", randomSeed);
						}
						startTime = SDL_GetTicks();
						dungeonGenerator.generateDungeon(screen, randomSeed, false);
						endTime = SDL_GetTicks();
						printf("finished %dx%d map in %.3f seconds\n", dungeonGenerator.getMapWidth(), dungeonGenerator.getMapHeight(), (endTime - startTime) / 1000.0);
						mapWidth = dungeonGenerator.getMapWidth();
						mapHeight = dungeonGenerator.getMapHeight();
						tileSize = screen->w / mapWidth;
						if (screen->h / mapHeight < tileSize)
							tileSize = screen->h / mapHeight;
						if (tileSize < MIN_TILE_SIZE)
							tileSize = MIN_TILE_SIZE;
						screenXOffset = (float)(screen->w - (mapWidth * tileSize)) / 2;
						screenYOffset = (float)(screen->h - (mapHeight * tileSize)) / 2;
						done = true;
					}
					else if (event.key.keysym.sym == SDLK_SPACE)
					{
						done = false;
						if (argc == 8)
						{
							randomSeed = (unsigned long)time(NULL);
							printf("Random seed: %lu\n", randomSeed);
						}
						startTime = SDL_GetTicks();
						dungeonGenerator.generateDungeon(screen, randomSeed, true);
						endTime = SDL_GetTicks();
						printf("finished %dx%d map in %.3f seconds\n", dungeonGenerator.getMapWidth(), dungeonGenerator.getMapHeight(), (endTime - startTime) / 1000.0);
						mapWidth = dungeonGenerator.getMapWidth();
						mapHeight = dungeonGenerator.getMapHeight();
						tileSize = screen->w / mapWidth;
						if (screen->h / mapHeight < tileSize)
							tileSize = screen->h / mapHeight;
						if (tileSize < MIN_TILE_SIZE)
							tileSize = MIN_TILE_SIZE;
						screenXOffset = (float)(screen->w - (mapWidth * tileSize)) / 2;
						screenYOffset = (float)(screen->h - (mapHeight * tileSize)) / 2;
						done = true;
					}
					else if (event.key.keysym.sym == SDLK_UP && done)
						upPressed = true;
					else if (event.key.keysym.sym == SDLK_DOWN && done)
						downPressed = true;
					else if (event.key.keysym.sym == SDLK_LEFT && done)
						leftPressed = true;
					else if (event.key.keysym.sym == SDLK_RIGHT && done)
						rightPressed = true;
					break;
				case SDL_KEYUP:
					if (event.key.keysym.sym == SDLK_UP && done)
						upPressed = false;
					else if (event.key.keysym.sym == SDLK_DOWN && done)
						downPressed = false;
					else if (event.key.keysym.sym == SDLK_LEFT && done)
						leftPressed = false;
					else if (event.key.keysym.sym == SDLK_RIGHT && done)
						rightPressed = false;
					break;
				case SDL_QUIT:
					quit = true;
					break;
			}
		}
		if (done)
		{
			startTime = SDL_GetTicks();
			dungeonGenerator.displayDungeon(screen, (int)screenXOffset, (int)screenYOffset);
			endTime = SDL_GetTicks();
			frameTime = (endTime - startTime) / 1000.0;
			if (upPressed)
				screenYOffset += moveSpeed * frameTime;
			else if (downPressed)
				screenYOffset -= moveSpeed * frameTime;
			if (leftPressed)
				screenXOffset += moveSpeed * frameTime;
			else if (rightPressed)
				screenXOffset -= moveSpeed * frameTime;
		}
	}
	return 0;
}
